/*
 * ke.h
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Jan 2006 - Created.
 */

/* 
 * ke.h:
 * Refered to ReactOS code
 */

#ifndef _KE_H
#define _KE_H

#include <linux/win32_thread.h>
#if 0
#include <linux/list.h>
#endif
#include "win32.h"

#ifdef CONFIG_UNIFIED_KERNEL
/* 
 * get_pre_mode() should be defined like: 
 * #define get_pre_mode() get_current_ethread()->tcb.previous_mode
 *
 * Temporarily, we define it as "UserMode" 
 */
#define	get_pre_mode()	UserMode

typedef enum _KOBJECTS {
   EventNotificationObject = 1,
   EventSynchronizationObject = 0,
   MutantObject = 2,
   ProcessObject = 3,
   QueueObject = 4,
   SemaphoreObject = 5,
   ThreadObject = 6,
   GateObject = 7,
   TimerNotificationObject = 8,
   TimerSynchronizationObject = 9,
   Spare2Object = 10,
   Spare3Object = 11,
   Spare4Object = 12,
   Spare5Object = 13,
   Spare6Object = 14,
   Spare7Object = 15,
   Spare8Object = 16,
   Spare9Object = 17,
   ApcObject = 18,
   DpcObject = 19,
   DeviceQueueObject = 20,
   EventPairObject = 21,
   InterruptObject = 22,
   ProfileObject = 23,
   ThreadedDpcObject = 24,
   MaximumKernelObject = 25
} KOBJECTS;

typedef LONG KPRIORITY;

static inline void enter_critical_region(void)
{
	struct ethread	*thread = current->ethread;

	if (thread)
		thread->tcb.kernel_apc_disable--;
}

static inline void leave_critical_region(void)
{
	struct ethread	*thread = current->ethread;

	/* FIXME: kernel apc check */
	if (thread && ++thread->tcb.kernel_apc_disable == 0) {
#if 0 
		if (list_empty(&CurrentThread->apc_state.apc_list_head[KernelMode]))
			KiKernelApcDeliveryCheck();
#endif
	}
}

static inline BOOLEAN
is_object_signaled(struct dispatcher_header *Object, struct kthread *Thread)
{ 
	/* Mutants are...well...mutants! */ 
	if (Object->type == MutantObject) { 
		/* Because Cutler hates mutants, they are actually signaled if the Signal State is <= 0 
		 * Well, only if they are recursivly acquired (i.e if we own it right now).  
		 * Of course, they are also signaled if their signal state is 1.  */ 
		if ((Object->signal_state <= 0 && ((struct kmutant *)Object)->owner_thread == Thread) 
				|| (Object->signal_state >= 1)) { 
			if(Object->signal_state >= 1) Object->signal_state =1;
			/* Signaled Mutant */ 
			return true; 
		} else { 
			/* Unsignaled Mutant */ 
			return false;
		} 
	} 

	/* Any other object is not a mutated freak, so let's use logic */ 
	return (!Object->signal_state <= 0);
}

extern void query_sys_time(large_integer_t *CurrentTime);

#endif /* CONFIG_UNIFIED_KERNEL */
#endif
