/*
 * process.h
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Jan 2006 - Created.
 */

/*
 * process.h:
 * Refered to Kernel-win32 code
 */
 
#ifndef _PROCESS_H
#define _PROCESS_H

#include <linux/module.h>
#include "win32_process.h"
#include <linux/binfmts.h>
#include "win32.h"
#include "object.h"
#include "ke.h"

#ifdef CONFIG_UNIFIED_KERNEL
#define	NtCurrentProcess()	((HANDLE)(ULONG_PTR)-1)
#define	PEB_BASE	0x7FFDF000

#define	MAX_PATH	256
#define	PROCESS_PRIO_NORMAL	6
#define	FREE_UNI(uni)	{ if (uni.Length) kfree(uni.Buffer); memset(&uni, 0, sizeof(uni)); }

#define NORMALIZE(x, addr)   if (x) x = (typeof(x))((void *)(x) + (void *)(addr))
#define DENORMALIZE(x, addr) if (x) x = (typeof(x))((void *)(x) - (void *)(addr))
#define	ALIGN_TO_LONG(x)	ALIGN((x), sizeof(LONG))
#define	PPF_NORMALIZED	1

#define	DENORMALIZE_PARAMS(params) \
{ \
	if ((params) && ((params)->Flags & PPF_NORMALIZED)) \
	{ \
		DENORMALIZE((params)->CurrentDirectoryName.Buffer, (params)); \
		DENORMALIZE((params)->DllPath.Buffer, (params)); \
		DENORMALIZE((params)->ImagePathName.Buffer, (params)); \
		DENORMALIZE((params)->CommandLine.Buffer, (params)); \
		DENORMALIZE((params)->WindowTitle.Buffer, (params)); \
		DENORMALIZE((params)->DesktopInfo.Buffer, (params)); \
		DENORMALIZE((params)->ShellInfo.Buffer, (params)); \
		DENORMALIZE((params)->RuntimeInfo.Buffer, (params)); \
		\
		(params)->Flags &= ~PPF_NORMALIZED; \
	} \
}

#define	NORMALIZE_PARAMS(params) \
{ \
	if ((params) && !((params)->Flags & PPF_NORMALIZED)) \
	{ \
		NORMALIZE((params)->CurrentDirectoryName.Buffer, (params)); \
		NORMALIZE((params)->DllPath.Buffer, (params)); \
		NORMALIZE((params)->ImagePathName.Buffer, (params)); \
		NORMALIZE((params)->CommandLine.Buffer, (params)); \
		NORMALIZE((params)->WindowTitle.Buffer, (params)); \
		NORMALIZE((params)->DesktopInfo.Buffer, (params)); \
		NORMALIZE((params)->ShellInfo.Buffer, (params)); \
		NORMALIZE((params)->RuntimeInfo.Buffer, (params)); \
		\
		(params)->Flags |= PPF_NORMALIZED; \
	} \
}

typedef VOID (STDCALL *PPEBLOCKROUTINE)(PVOID);

typedef struct _PEB_LDR_DATA
{
    ULONG               Length;
    BOOLEAN             Initialized;
    PVOID               SsHandle;
    LIST_ENTRY          InLoadOrderModuleList;
    LIST_ENTRY          InMemoryOrderModuleList;
    LIST_ENTRY          InInitializationOrderModuleList;
    PVOID               EntryInProgress;
} PEB_LDR_DATA, *PPEB_LDR_DATA;

typedef struct RTL_DRIVE_LETTER_CURDIR
{
	USHORT              Flags;
	USHORT              Length;
	ULONG               TimeStamp;
	UNICODE_STRING      DosPath;
} RTL_DRIVE_LETTER_CURDIR, *PRTL_DRIVE_LETTER_CURDIR;

typedef struct _RTL_USER_PROCESS_PARAMETERS
{
    ULONG  AllocationSize;
    ULONG  Size;
    ULONG  Flags;
    ULONG  DebugFlags;
    HANDLE  hConsole;
    ULONG  ProcessGroup;
    HANDLE  hStdInput;
    HANDLE  hStdOutput;
    HANDLE  hStdError;
    UNICODE_STRING  CurrentDirectoryName;
    HANDLE  CurrentDirectoryHandle;
    UNICODE_STRING  DllPath;
    UNICODE_STRING  ImagePathName;
    UNICODE_STRING  CommandLine;
    PWSTR  Environment;
    ULONG  dwX;
    ULONG  dwY;
    ULONG  dwXSize;
    ULONG  dwYSize;
    ULONG  dwXCountChars;
    ULONG  dwYCountChars;
    ULONG  dwFillAttribute;
    ULONG  dwFlags;
    ULONG  wShowWindow;
    UNICODE_STRING  WindowTitle;
    UNICODE_STRING  DesktopInfo;
    UNICODE_STRING  ShellInfo;
    UNICODE_STRING  RuntimeInfo;
    RTL_DRIVE_LETTER_CURDIR DLCurrentDirectory[0x20];
} RTL_USER_PROCESS_PARAMETERS, *PRTL_USER_PROCESS_PARAMETERS;

typedef struct _PEB_FREE_BLOCK
{
    struct _PEB_FREE_BLOCK* Next;
    ULONG Size;
} PEB_FREE_BLOCK, *PPEB_FREE_BLOCK;

typedef struct _PEB
{
	UCHAR 				InheritedAddressSpace;		/* 00h */
	UCHAR 				ReadImageFileExecOptions;	/* 01h */
	UCHAR 				BeingDebugged;			/* 02h */
	BOOLEAN 			SpareBool;			/* 03h */
	HANDLE 				Mutant;				/* 04h */
	PVOID 				ImageBaseAddress;		/* 08h */
	PPEB_LDR_DATA 			Ldr;				/* 0Ch */
	PRTL_USER_PROCESS_PARAMETERS 	ProcessParameters;		/* 10h */
	PVOID 				SubSystemData;			/* 14h */
	PVOID 				ProcessHeap;			/* 18h */
	PVOID 				FastPebLock;			/* 1Ch */
	PPEBLOCKROUTINE 		FastPebLockRoutine;		/* 20h */
	PPEBLOCKROUTINE 		FastPebUnlockRoutine;		/* 24h */
	ULONG 				EnvironmentUpdateCount;		/* 28h */
	PVOID* 				KernelCallbackTable;		/* 2Ch */
	PVOID 				EventLogSection;		/* 30h */
	PVOID 				EventLog;			/* 34h */
	PPEB_FREE_BLOCK 		FreeList;			/* 38h */
	ULONG 				TlsExpansionCounter;		/* 3Ch */
	PVOID 				TlsBitmap;			/* 40h */
	ULONG 				TlsBitmapBits[0x2];		/* 44h */
	PVOID 				ReadOnlySharedMemoryBase;	/* 4Ch */
	PVOID 				ReadOnlySharedMemoryHeap;	/* 50h */
	PVOID* 				ReadOnlyStaticServerData;	/* 54h */
	PVOID 				AnsiCodePageData;		/* 58h */
	PVOID 				OemCodePageData;		/* 5Ch */
	PVOID 				UnicodeCaseTableData;		/* 60h */
	ULONG 				NumberOfProcessors;		/* 64h */
	ULONG 				NtGlobalFlag;			/* 68h */
	BYTE                            Spare2[4];                      /* 6ch */
	LARGE_INTEGER 			CriticalSectionTimeout;		/* 70h */
	ULONG 				HeapSegmentReserve;		/* 78h */
	ULONG 				HeapSegmentCommit;		/* 7Ch */
	ULONG 				HeapDeCommitTotalFreeThreshold;	/* 80h */
	ULONG 				HeapDeCommitFreeBlockThreshold;	/* 84h */
	ULONG 				NumberOfHeaps;			/* 88h */
	ULONG 				MaximumNumberOfHeaps;		/* 8Ch */
	PVOID* 				ProcessHeaps;			/* 90h */
	PVOID 				GdiSharedHandleTable;		/* 94h */
	PVOID 				ProcessStarterHelper;		/* 98h */
	PVOID 				GdiDCAttributeList;		/* 9Ch */
	PVOID 				LoaderLock;			/* A0h */
	ULONG 				OSMajorVersion;			/* A4h */
	ULONG 				OSMinorVersion;			/* A8h */
	USHORT 				OSBuildNumber;			/* ACh */
	USHORT 				OSCSDVersion;			/* AEh */
	ULONG 				OSPlatformId;			/* B0h */
	ULONG 				ImageSubSystem;			/* B4h */
	ULONG 				ImageSubSystemMajorVersion;	/* B8h */
	ULONG 				ImageSubSystemMinorVersion;	/* BCh */
	ULONG 				ImageProcessAffinityMask;	/* C0h */
	ULONG 				GdiHandleBuffer[0x22];		/* C4h */
	PVOID 				PostProcessInitRoutine;		/* 14Ch */
	PVOID 				*TlsExpansionBitmap;		/* 150h */
	ULONG 				TlsExpansionBitmapBits[0x20];	/* 154h */
	ULONG 				SessionId;			/* 1D4h */
	PVOID 				AppCompatInfo;			/* 1D8h */
	UNICODE_STRING 			CSDVersion;			/* 1DCh */
} PEB, *PPEB;

/* process startup state */
enum startup_state { STARTUP_IN_PROGRESS, STARTUP_DONE, STARTUP_ABORTED };
#ifdef SERVER_PM
struct w32process
{
    struct w32process    *parent;          /* parent process */
    struct eprocess      *eprocess;
    struct w32thread     *debugger;        /* thread debugging this process */
    process_id_t         group_id;        /* group id of the process */
    struct timeout_user *sigkill_timeout; /* timeout for final SIGKILL */
    int                  running_threads; /* number of threads running in this process */
    timeout_t			start_time;      /* absolute time at process start */
    timeout_t           end_time;        /* absolute time at process end */
    int                  suspend;         /* global process suspend count */
    int                  is_system;       /* is it a system process? */
    unsigned int         create_flags;    /* process creation flags */
    struct list_head     locks;           /* list of file locks owned by the process */
    struct list_head     classes;         /* window classes owned by the process */
    struct console_input*console;         /* console input */
    enum startup_state   startup_state;   /* startup state */
    struct startup_info *startup_info;    /* startup info while init is in progress */
    struct kevent       *idle_event;      /* event for input idle */
    struct msg_queue    *queue;           /* main message queue */
    obj_handle_t         winstation;      /* main handle to process window station */
    obj_handle_t         desktop;         /* handle to desktop to use for new threads */
    struct list_head     dlls;            /* list of loaded dlls */
    void                *ldt_copy;        /* pointer to LDT copy in client addr space */
    unsigned int         trace_data;      /* opaque data used by the process tracing mechanism */
};
#else
struct w32process
{
    struct object        obj;             /* object header */
    struct list_head          entry;           /* entry in system-wide process list */
    struct w32process    *parent;          /* parent process */
    struct eprocess      *eprocess;
    struct list_head          thread_list;     /* thread list */
    struct w32thread     *debugger;        /* thread debugging this process */
    struct handle_table *handles;         /* handle entries */
    process_id_t         id;              /* id of the process */
    process_id_t         group_id;        /* group id of the process */
    struct timeout_user *sigkill_timeout; /* timeout for final SIGKILL */
    int                  unix_pid;        /* Unix pid for final SIGKILL */
    int                  exit_code;       /* process exit code */
    int                  running_threads; /* number of threads running in this process */
    timeout_t		start_time;      /* absolute time at process start */
    timeout_t           end_time;        /* absolute time at process end */
    int                  priority;        /* priority class */
    int                  affinity;        /* process affinity mask */
    int                  suspend;         /* global process suspend count */
    int                  is_system;       /* is it a system process? */
    unsigned int         create_flags;    /* process creation flags */
    struct list_head          locks;           /* list of file locks owned by the process */
    struct list_head          classes;         /* window classes owned by the process */
    struct console_input*console;         /* console input */
    enum startup_state   startup_state;   /* startup state */
    struct startup_info *startup_info;    /* startup info while init is in progress */
    struct kevent       *idle_event;      /* event for input idle */
    struct msg_queue    *queue;           /* main message queue */
    obj_handle_t         winstation;      /* main handle to process window station */
    obj_handle_t         desktop;         /* handle to desktop to use for new threads */
    struct token        *token;           /* security token associated with this process */
    struct list_head          dlls;            /* list of loaded dlls */
    void                *peb;             /* PEB address in client address space */
    void                *ldt_copy;        /* pointer to LDT copy in client addr space */
    unsigned int         trace_data;      /* opaque data used by the process tracing mechanism */
};
#endif

extern POBJECT_TYPE process_object_type;
void kprocess_init(struct kprocess *process, char prio,
		unsigned long affinity, physical_address_t dir_table_base);
void eprocess_init(struct eprocess *process);
NTSTATUS STDCALL create_peb(struct eprocess *process);
void exit_process_threads(struct eprocess *process, NTSTATUS ExitStatus);
void exit_current_thread(struct task_struct *tsk, NTSTATUS ExitStatus);
VOID delete_process(PVOID Object);
int poll_process(struct wait_table_entry *wte);
int lookup_process_by_pid(HANDLE pid, struct eprocess** process);
NTSTATUS lookup_process_thread_by_cid(PCLIENT_ID cid, PEPROCESS *process, PETHREAD *thread);

void __exit_process(struct eprocess * process);

/* for server process/thread management migration */
struct process_dll
{
    struct list_head          entry;           /* entry in per-process dll list */
    struct file         *file;            /* dll file */
    void                *base;            /* dll base address (in process addr space) */
    size_t               size;            /* dll size */
    void                *name;            /* ptr to ptr to name (in process addr space) */
    int                  dbg_offset;      /* debug info offset */
    int                  dbg_size;        /* debug info size */
    data_size_t          namelen;         /* length of dll file name */
    WCHAR               *filename;        /* dll file name */
};

struct process_snapshot
{
    struct w32process *process;  /* process ptr */
    int             count;    /* process refcount */
    int             threads;  /* number of threads */
    int             priority; /* priority class */
    int             handles;  /* number of handles */
};

struct module_snapshot
{
    void           *base;     /* module base addr */
    size_t          size;     /* module size */
    data_size_t     namelen;  /* length of file name */
    WCHAR          *filename; /* module file name */
};

extern void enum_processes( int (*cb)(struct w32process*, void*), void *user);
extern struct process_snapshot *process_snap( int *count );
extern struct module_snapshot *module_snap( struct w32process *process, int *count );
/* for server process/thread management migration */

static inline struct eprocess *process2eprocess(struct w32process *proc)
{
    return proc->eprocess;
}

static inline struct w32process *eprocess2process(struct eprocess *eproc)
{
    return (struct w32process *)eproc->win32process;
}

static inline struct eprocess *get_current_eprocess(void)
{
	return current->ethread ? current->ethread->threads_process : NULL;
}

static inline struct w32process *get_current_w32process(void)
{
    return current->ethread ? current->ethread->threads_process->win32process : NULL;
}

static inline void lock_process(struct eprocess *process)
{
	enter_critical_region();
	spin_lock(&process->process_lock);
}

static inline void unlock_process(struct eprocess *process)
{
	spin_unlock(&process->process_lock);
	leave_critical_region();
}

static inline unsigned int get_process_id( struct w32process *process ) { 
	return (unsigned int)process2eprocess(process)->unique_processid; 
}

static inline int is_process_init_done( struct w32process *process )
{
    return process->startup_state == STARTUP_DONE;
}

static inline struct process_dll *get_process_exe_module( struct w32process *process )
{
    struct list_head *ptr = list_head( &process->dlls );
    return ptr ? LIST_ENTRY( ptr, struct process_dll, entry ) : NULL;
}
NTSTATUS create_ppb(PRTL_USER_PROCESS_PARAMETERS *ppb_res,
		struct eprocess *process,
		struct linux_binprm *bprm,
		char *image_name,
		char *dll_path,
		char *current_dir,
		PWSTR environ,
		char *window_title,
		char *desktop_info,
		char *shell_info,
		char *rt_info);

NTSTATUS SERVICECALL
NtTerminateProcess(IN HANDLE ProcessHandle  OPTIONAL,
                   IN NTSTATUS ExitStatus);

long init_cid_table(void);
void destroy_cid_table(void);
HANDLE create_cid_handle(PVOID object, POBJECT_TYPE obj_type);
int delete_cid_handle(HANDLE cid_handle, POBJECT_TYPE obj_type);

int flush_old_exec_from_task(struct task_struct *task);

#endif /* CONFIG_UNIFIED_KERNEL */
#endif /* _PROCESS_H */
