/*
 * uk_lib.h
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Mar 2009 - Created.
 */

/* 
 * uk_lib.h:
 * Refered to Wine code
 */

#ifndef _WINESERVER_UK_LIB_H
#define _WINESERVER_UK_LIB_H
#include <linux/mm.h>
#include <linux/syscalls.h>
#include <linux/file.h>
#include <linux/vmalloc.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/statfs.h>
#if 0
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/fs_struct.h>
#include <linux/dcache.h>
#endif

/* FIXME */
#include "win32.h"
#include "w32syscall.h"
#include "handle.h"
#include "event.h"

#include "server.h"
#include "request.h"
#include "thread.h"
#include "winerror.h"
#include "object.h"
#include "objwait.h"
#include "user.h"
#include "file.h"
#include "wcstr.h"

#ifdef CONFIG_UNIFIED_KERNEL
#define	errno	get_error()

struct LIBC_FILE
{
  struct file *filp;
  loff_t pos;
  char *buf;
  long buflen;
  long bufpos;
  ssize_t validlen;
};

#if 0
struct pipe_server
{
  int fd;
  /* Depends on how pipe is implemented */
};
#endif

time_t time(void* v);

#define MAXSIZE_ALLOC (128 * 1024)

void *malloc(size_t size);
void* calloc(size_t nmemb, size_t size);
void *mem_alloc(size_t size);
void *memdup(const void *data, size_t len);
void free(void* p);
void *realloc(void *ptr, size_t new_size, size_t old_size);

WCHAR toupperW(WCHAR ch);
WCHAR tolowerW(WCHAR ch);
int memicmpW(const WCHAR *str1, const WCHAR *str2, int n);

void no_flush(struct fd *fd, struct kevent **event);
int no_add_queue(struct object *obj, struct wait_queue_entry *entry);
int no_close_handle(struct object *o, struct w32process *p, obj_handle_t h);
struct object *no_lookup_name(struct object *obj, struct unicode_str *name, unsigned int attributes);
struct fd *no_get_fd(struct object *obj);
struct object *no_open_file(struct object *o, unsigned int d , unsigned int i , unsigned int u);
int no_satisfied(struct object *obj, struct w32thread *thread);
int no_signal(struct object *obj, unsigned int access);
unsigned int no_map_access(struct object *obj, unsigned int access);

/* object */
#if 0
POBJECT_ATTRIBUTES get_attr(HANDLE RootDirectory, const struct unicode_str *name,
		unsigned int attr, ULONG len, void *securitydes, void *securityqofs);

void *create_named_object(HANDLE RootDirectory, POBJECT_TYPE Type,
		const struct unicode_str *name, unsigned int attr, int obj_size);
#endif

void dump_object_name(struct object *obj);

HANDLE duplicate_handle(HANDLE src, HANDLE src_handle, HANDLE dst,
		unsigned int access, unsigned int attr, unsigned int options);

#define HANDLE_FLAG_INHERIT             0x00000001

#define HANDLE_FLAG_PROTECT_FROM_CLOSE  0x00000002

/* reserved handle access rights */

#define RESERVED_SHIFT         26

#define RESERVED_INHERIT       (HANDLE_FLAG_INHERIT << RESERVED_SHIFT)

#define RESERVED_CLOSE_PROTECT (HANDLE_FLAG_PROTECT_FROM_CLOSE << RESERVED_SHIFT)

#define RESERVED_ALL           (RESERVED_INHERIT | RESERVED_CLOSE_PROTECT)

unsigned int get_handle_access(struct eprocess *process, HANDLE handle);

int close_handle(struct eprocess *process, HANDLE handle);

extern long uk_close(unsigned int fd);
extern ssize_t uk_write(unsigned int fd, void *buf, size_t size);
extern ssize_t uk_filp_write(struct file *filp, void *buf, size_t size);
extern long uk_fcntl(unsigned int fd, int cmd, unsigned long arg /* struct flock *flock */);
extern long uk_fcntl_lock(unsigned int fd, int op, struct flock* flock);
extern long uk_dup(unsigned int fd);

extern ssize_t uk_pread(unsigned int fd, char *buf, size_t count, off_t pos);
extern ssize_t uk_filp_pread(struct file *filp, char *buf, size_t count, off_t pos);
extern ssize_t uk_pwrite(unsigned int fd, const char *buf, size_t count, off_t pos);
extern ssize_t uk_filp_pwrite(struct file *filp, const char *buf, size_t count, off_t pos);
extern long uk_readlink(const char *path, char *buf, size_t bufsiz);
extern long uk_fcntl(unsigned int fd, int cmd, unsigned long arg);
extern long uk_stat(char *filename, struct stat *st);

/* fd */
void remove_timeout_user(struct timeout_user *user);

/* FIXME fd not implemented */
struct fd *create_anonymous_fd(const struct fd_ops *fd_user_ops,
		int unix_fd, struct object *user, unsigned int options);
void *get_fd_user(struct fd *fd);
int get_unix_fd(HANDLE handle);
#if 0
int get_file_unix_fd(struct  file* f);
struct file *get_file_obj(struct w32process *p, user_handle_t h, unsigned int d);
#endif

/* winstation */
obj_handle_t find_inherited_handle(struct w32process *process, const struct object_ops *ops);
void cleanup_clipboard_thread(struct w32thread *thread);
void msg_queue_dump(struct object *obj, int verbose);
/* console */
void kill_console_processes(struct w32thread *renderer, int exit_code);

void get_req_path(struct unicode_str *str, int skip_root);

/* add by solaris 2008/06/23 */
int vsnprintfW(WCHAR *str, size_t len, const WCHAR *format, va_list valist);

int vsprintfW(WCHAR *str, const WCHAR *format, va_list valist);

int snprintfW(WCHAR *str, size_t len, const WCHAR *format, ...);

extern int sprintfW(WCHAR *str, const WCHAR *format, ...);

extern void *uk_fgets(void *buf, int len, struct LIBC_FILE *fp);


extern long uk_fclose(struct LIBC_FILE* fp);

extern struct LIBC_FILE *uk_fdopen(int fd, char *readwrite);
extern struct LIBC_FILE *uk_fopen(const char *filename, char *readwrite) ;
extern struct LIBC_FILE *uk_filp_open(struct file *filp, char *readwrite) ;

extern void uk_perror(const char *s);
extern int uk_fprintf(struct LIBC_FILE *fp , char *fmt, ...);
extern ssize_t uk_fwrite(struct LIBC_FILE *fp, void *buf, size_t size);

extern long uk_lstat(char *filename, struct stat *st);
extern long uk_fstatfs(unsigned int fd, struct statfs *st);
extern int uk_realpath(const char *path, char *real);
extern long uk_fstat(unsigned int fd, struct stat *st);
extern long uk_ftruncate(unsigned int fd, long tr);
extern long uk_open(const char *filename, int flags, int mode);
extern void uk_unlink(const char *filename);
extern int uk_rename(const char *oldpath, const char *newpath);
extern ssize_t uk_read(unsigned int fd, void *buf, size_t size);
extern long uk_mkdir(const char *name, int mode);

extern POBJECT_TYPE event_object_type;

static inline void set_error(unsigned int e)
{
	struct w32thread	*w32thread = get_current_w32thread();

	if (w32thread)
		w32thread->error = e;
}

static inline unsigned int get_error(void)
{
	struct w32thread	*w32thread = get_current_w32thread();

	return w32thread ? w32thread->error : 0;
}

static inline void clear_error(void)
{
	set_error(STATUS_SUCCESS);
}

static inline pid_t getpid(void)
{
	return current->tgid;
}

static inline char *strdup(const char *src)
{
	char	*ret;

	ret = kstrdup(src, GFP_KERNEL);
	if (!ret)
		set_error(STATUS_NO_MEMORY);

	return ret;
}

static inline void *grab_object(void *obj)
{
	ref_object(obj);
	return obj;
}

static inline void *get_handle_obj(HANDLE handle, unsigned int access, POBJECT_TYPE Type)
{
	void	*obj = NULL;
	NTSTATUS	ret;

	ret = ref_object_by_handle(handle, access, Type, KernelMode, &obj, NULL);
	if (!NT_SUCCESS(ret)) {
		set_error((unsigned int)ret);
		return NULL;
	}

	return obj;
}

static inline void *get_wine_handle_obj(struct w32process *proc, HANDLE handle, 
				unsigned int access, const struct object_ops *ops)
{
	void	*obj = NULL;
	NTSTATUS	ret;

	ret = ref_object_by_handle(handle, access, NULL, KernelMode, &obj, NULL);
	if (!NT_SUCCESS(ret)) {
		set_error((unsigned int)ret);
		return NULL;
	}

	return obj;
}

static inline HANDLE alloc_key_handle(void *key, unsigned int access, unsigned int attr)
{
	HANDLE handle = NULL;
	NTSTATUS ret;

	ret = create_handle(current->ethread ? get_current_eprocess() : NULL,
			key, access, attr & OBJ_INHERIT, &handle);
	if (!NT_SUCCESS(ret)) {
		set_error((unsigned int)ret);
		return NULL;
	}

	return handle;
}

static inline void *get_req_data(void)
{
	return get_current_w32thread() ? get_current_w32thread()->req_data : NULL;
}

static inline unsigned int get_req_data_size(void)
{
	return get_current_w32thread() ? get_current_w32thread()->req.request_header.request_size : 0;
}

static inline unsigned int get_reply_max_size(void)
{
	return get_current_w32thread() ? get_current_w32thread()->req.request_header.reply_size : 0;
}

static inline void set_reply_data_ptr(void *data, unsigned int size)
{
	struct w32thread *thread = get_current_w32thread();

	if (!thread)
		return;

	if (size <= get_reply_max_size()) {
		thread->reply_size = size;
		thread->reply_data = data;
	} else
		kdebug("size large err\n"); /* TODO */
}

static inline void *set_reply_data_size(int size)
{
	struct w32thread *thread = get_current_w32thread();

	if (!thread)
		return NULL;

	if (size <= get_reply_max_size()) {
		if (size && !(thread->reply_data = mem_alloc(size)))
			size = 0;
		thread->reply_size = size;
		return thread->reply_data;

	} else {
		kdebug("size large err\n"); /* TODO */
		return NULL;
	}
}

static inline void *set_reply_data(const void *data, unsigned int size)
{
	void *ret = set_reply_data_size(size);

	if (ret)
		memcpy(ret, data, size);

	return ret;
}

static inline void get_req_unicode_str(struct unicode_str *s)
{
	s->str = get_req_data();
	s->len = (get_req_data_size() / sizeof(WCHAR)) * sizeof(WCHAR);
}

static inline int uk_fputc(char c, struct LIBC_FILE *f)
{
	return uk_fprintf(f,"%c",c);;
}

static inline WCHAR *strcpyW(WCHAR *dst, const WCHAR *src)
{
	return wcscpy(dst, src);
}

static inline int strlenW(const WCHAR *str)
{
	return wcslen((PWSTR)str) / sizeof(WCHAR);
}

static inline WCHAR *memchrW(const WCHAR *ptr, WCHAR ch, size_t n)
{
	return wmemchr(ptr, ch, n);
}

static inline int strncmpW(const WCHAR *str1, const WCHAR *str2, int n)
{
	return wcsncmp(str1, str2, n);
}

static inline int isdigitW(WCHAR wc)
{
	return (wc >= '0') && (wc <= '9');
}

static inline struct kevent *get_event_obj(struct w32process *process, obj_handle_t handle, 
		unsigned int access)
{
	return (struct kevent *)get_handle_obj(handle, access, event_object_type);
}

static inline void set_win32_error(int e)
{
	set_error(0xc0010000 | e);
}

static inline void *get_ptid_entry(HANDLE pid)
{
	NTSTATUS ret;
	struct eprocess *process = NULL;

	ret = lookup_process_by_pid(pid, &process);
	if (!NT_SUCCESS(ret))
		set_error((unsigned int)ret);

	return process;
}

static inline struct eprocess *get_process_from_id(HANDLE id)
{
	return get_ptid_entry(id);
}

static inline void free_ptid(HANDLE pid)
{
	NTSTATUS ret;

	ret = delete_cid_handle(pid, process_object_type);
	if (!NT_SUCCESS(ret))
		set_error((unsigned int)ret);

	return;
}

static inline int send_thread_signal(struct ethread *thread, int sig)
{
	return send_sig_info(sig, (void *)2L, thread->et_task);
}

static inline WCHAR *get_object_name(void *obj, data_size_t *len)
{
	POBJECT_HEADER_NAME_INFO name_info;

	name_info = HEADER_TO_OBJECT_NAME(BODY_TO_HEADER(obj));
	if (name_info) {
		*len = name_info->Name.Length;
		return name_info->Name.Buffer;
	} else {
		*len = 0;
		return NULL;
	}
}

static inline int interlocked_xchg(int *dest, int val)
{
	int ret;

	asm volatile("lock; xchgl %0,(%1)"
			: "=r" (ret)
			: "r" (dest), "0" (val)
			: "memory");

	return ret;
}
static inline int interlocked_xchg_add(int *dest, int incr)
{
	int ret;

	asm volatile("lock; xaddl %0,(%1)"
			: "=r" (ret)
			: "r" (dest), "0" (incr)
			: "memory" );

	return ret;
}

static inline struct inode *get_file_inode(struct file *file)
{
    return file->f_path.dentry->d_inode;
}

static inline loff_t get_file_size(struct file *file)
{
    return file->f_path.dentry->d_inode->i_size;
}

enum
{
  SHUT_RD = 0,      /* No more receptions.  */
#define SHUT_RD     SHUT_RD
  SHUT_WR,      /* No more transmissions.  */
#define SHUT_WR     SHUT_WR
  SHUT_RDWR     /* No more receptions or transmissions.  */
#define SHUT_RDWR   SHUT_RDWR
};

#define POLLIN      0x001       /* There is data to read.  */
#define POLLPRI     0x002       /* There is urgent data to read.  */
#define POLLOUT     0x004       /* Writing now will not block.  */
#define POLLERR     0x008       /* Error condition.  */
#define POLLHUP     0x010       /* Hung up.  */
#define POLLNVAL    0x020       /* Invalid polling request.  */

/* Data structure describing a polling request.  */
struct pollfd
{
    int fd;         /* File descriptor to poll.  */
    short int events;       /* Types of events poller cares about.  */
    short int revents;      /* Types of events that actually occurred.  */
};
#endif
#endif /* _WINESERVER_UK_LIB_H */
